from mdk.types.specifier import TypeSpecifier

__all__ = ['AnyType', 'IntType', 'FloatType', 'BoolType', 'ShortType', 'ByteType', 'CharType', 'StringType', 'StateNoType', 'AnimType', 'SoundType', 'SpriteType']

AnyType: TypeSpecifier
IntType: TypeSpecifier
FloatType: TypeSpecifier
BoolType: TypeSpecifier
ShortType: TypeSpecifier
ByteType: TypeSpecifier
CharType: TypeSpecifier
StringType: TypeSpecifier
UStringType: TypeSpecifier
StateNoType: TypeSpecifier
AnimType: TypeSpecifier
SoundType: TypeSpecifier
SpriteType: TypeSpecifier
