from mdk.types.specifier import TypeSpecifier as TypeSpecifier

def check_types_assignable(spec1: TypeSpecifier, spec2: TypeSpecifier) -> TypeSpecifier | None: ...
def check_any_assignable(spec1: TypeSpecifier, spec2: list[TypeSpecifier]) -> TypeSpecifier | None: ...
