from mdk.types.builtins import BoolType as BoolType
from mdk.types.context import CompilerContext as CompilerContext
from mdk.types.expressions import Expression as Expression
from mdk.types.variables import VariableExpression as VariableExpression
from mdk.utils.expressions import check_types_assignable as check_types_assignable
from mdk.utils.shared import format_bool as format_bool

def TriggerAnd(expr1: Expression, expr2: Expression): ...
def TriggerOr(expr1: Expression, expr2: Expression): ...
def TriggerNot(expr: Expression): ...
def TriggerAssign(expr1: Expression, expr2: Expression): ...
def TriggerPush(): ...
def TriggerPop(): ...
