from mdk.types.builtins import *
from mdk.types.defined import *
from mdk.types.context import StateController
from mdk.types.expressions import ConvertibleExpression, Expression, TupleExpression
from mdk.resources.animation import Animation
from typing import Callable

__all__ = ['AfterImage', 'AfterImageTime', 'AngleAdd', 'AngleDraw', 'AngleMul', 'AngleSet', 'AssertSpecial', 'AttackDist', 'AttackMulSet', 'BindToParent', 'BindToRoot', 'BindToTarget', 'ChangeAnim', 'ChangeAnim2', 'ChangeState', 'ClearClipboard', 'CtrlSet', 'DefenceMulSet', 'DestroySelf', 'EnvColor', 'EnvShake', 'Explod', 'ExplodBindTime', 'ForceFeedback', 'FallEnvShake', 'GameMakeAnim', 'Gravity', 'Helper', 'HitAdd', 'HitBy', 'HitDef', 'HitFallDamage', 'HitFallSet', 'HitFallVel', 'HitOverride', 'HitVelSet', 'LifeAdd', 'LifeSet', 'MakeDust', 'ModifyExplod', 'MoveHitReset', 'NotHitBy', 'Null', 'Offset', 'PalFX', 'ParentVarAdd', 'ParentVarSet', 'Pause', 'PlayerPush', 'PlaySnd', 'PosAdd', 'PosFreeze', 'PosSet', 'PowerAdd', 'PowerSet', 'Projectile', 'RemapPal', 'RemoveExplod', 'ReversalDef', 'ScreenBound', 'SelfState', 'SprPriority', 'StateTypeSet', 'SndPan', 'SuperPause', 'TargetBind', 'TargetDrop', 'TargetFacing', 'TargetLifeAdd', 'TargetPowerAdd', 'TargetState', 'TargetVelAdd', 'TargetVelSet', 'Trans', 'Turn', 'VarAdd', 'VarSet', 'VarRandom', 'VarRangeSet', 'VelAdd', 'VelMul', 'VelSet', 'VictoryQuote', 'Width', 'BGPalFX', 'AllPalFX', 'AppendToClipboard', 'DisplayToClipboard']

def AfterImage(time: ConvertibleExpression | None = None, length: ConvertibleExpression | None = None, palcolor: ConvertibleExpression | None = None, palinvertall: ConvertibleExpression | None = None, palbright: TupleExpression | None = None, palcontrast: TupleExpression | None = None, palpostbright: TupleExpression | None = None, paladd: TupleExpression | None = None, palmul: TupleExpression | None = None, timegap: ConvertibleExpression | None = None, framegap: ConvertibleExpression | None = None, trans: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AfterImageTime(time: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AllPalFX(time: ConvertibleExpression | None = None, add: TupleExpression | None = None, mul: TupleExpression | None = None, sinadd: TupleExpression | None = None, invertall: ConvertibleExpression | None = None, color: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AngleAdd(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AngleDraw(value: ConvertibleExpression | None = None, scale: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AngleMul(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AngleSet(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AppendToClipboard(text: ConvertibleExpression, params: TupleExpression | None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AssertSpecial(flag: ConvertibleExpression, flag2: ConvertibleExpression | None = None, flag3: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AttackDist(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def AttackMulSet(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def BGPalFX(time: ConvertibleExpression | None = None, add: TupleExpression | None = None, mul: TupleExpression | None = None, sinadd: TupleExpression | None = None, invertall: ConvertibleExpression | None = None, color: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def BindToParent(time: ConvertibleExpression | None = None, facing: ConvertibleExpression | None = None, pos: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def BindToRoot(time: ConvertibleExpression | None = None, facing: ConvertibleExpression | None = None, pos: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def BindToTarget(time: ConvertibleExpression | None = None, id: ConvertibleExpression | None = None, pos: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ChangeAnim(value: ConvertibleExpression | Animation, elem: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ChangeAnim2(value: ConvertibleExpression | Animation, elem: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ChangeState(value: ConvertibleExpression | str | int | Callable[..., None | StateController], ctrl: ConvertibleExpression | None = None, anim: ConvertibleExpression | Animation | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ClearClipboard(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def CtrlSet(ctrl: ConvertibleExpression | None = None, value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def DefenceMulSet(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def DestroySelf(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def DisplayToClipboard(text: ConvertibleExpression, params: TupleExpression | None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def EnvColor(value: TupleExpression | None = None, time: ConvertibleExpression | None = None, under: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def EnvShake(time: ConvertibleExpression, freq: ConvertibleExpression | None = None, ampl: ConvertibleExpression | None = None, phase: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Explod(anim: ConvertibleExpression | Animation, id: ConvertibleExpression | None = None, pos: TupleExpression | None = None, postype: ConvertibleExpression | None = None, space: ConvertibleExpression | None = None, facing: ConvertibleExpression | None = None, vfacing: ConvertibleExpression | None = None, bindtime: ConvertibleExpression | None = None, vel: TupleExpression | None = None, accel: TupleExpression | None = None, random: TupleExpression | None = None, removetime: ConvertibleExpression | None = None, supermove: ConvertibleExpression | None = None, supermovetime: ConvertibleExpression | None = None, pausemovetime: ConvertibleExpression | None = None, scale: TupleExpression | None = None, sprpriority: ConvertibleExpression | None = None, ontop: ConvertibleExpression | None = None, shadow: ConvertibleExpression | None = None, ownpal: ConvertibleExpression | None = None, removeongethit: ConvertibleExpression | None = None, trans: ConvertibleExpression | None = None, angle: ConvertibleExpression | None = None, alpha: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ExplodBindTime(id: ConvertibleExpression | None = None, time: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ForceFeedback(waveform: ConvertibleExpression | None = None, time: ConvertibleExpression | None = None, freq: TupleExpression | None = None, ampl: TupleExpression | None = None, self: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def FallEnvShake(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def GameMakeAnim(value: ConvertibleExpression | None = None, under: ConvertibleExpression | None = None, pos: TupleExpression | None = None, random: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Gravity(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Helper(helpertype: ConvertibleExpression | None = None, name: ConvertibleExpression | None = None, id: ConvertibleExpression | None = None, pos: TupleExpression | None = None, postype: ConvertibleExpression | None = None, facing: ConvertibleExpression | None = None, stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, keyctrl: ConvertibleExpression | None = None, ownpal: ConvertibleExpression | None = None, supermovetime: ConvertibleExpression | None = None, pausemovetime: ConvertibleExpression | None = None, size_xscale: ConvertibleExpression | None = None, size_yscale: ConvertibleExpression | None = None, size_ground_back: ConvertibleExpression | None = None, size_ground_front: ConvertibleExpression | None = None, size_air_back: ConvertibleExpression | None = None, size_air_front: ConvertibleExpression | None = None, size_height: ConvertibleExpression | None = None, size_proj_doscale: ConvertibleExpression | None = None, size_head_pos: TupleExpression | None = None, size_mid_pos: TupleExpression | None = None, size_shadowoffset: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitAdd(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitBy(value: TupleExpression | None = None, value2: TupleExpression | None = None, time: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitDef(attr: TupleExpression, hitflag: ConvertibleExpression | None = None, guardflag: ConvertibleExpression | None = None, affectteam: ConvertibleExpression | None = None, animtype: ConvertibleExpression | None = None, air_animtype: ConvertibleExpression | None = None, fall_animtype: ConvertibleExpression | None = None, priority: TupleExpression | None = None, damage: ConvertibleExpression | None = None, pausetime: TupleExpression | None = None, guard_pausetime: TupleExpression | None = None, sparkno: ConvertibleExpression | None = None, guard_sparkno: ConvertibleExpression | None = None, sparkxy: TupleExpression | None = None, hitsound: TupleExpression | None = None, guardsound: TupleExpression | None = None, ground_type: ConvertibleExpression | None = None, air_type: ConvertibleExpression | None = None, ground_slidetime: ConvertibleExpression | None = None, guard_slidetime: ConvertibleExpression | None = None, ground_hittime: ConvertibleExpression | None = None, guard_hittime: ConvertibleExpression | None = None, air_hittime: ConvertibleExpression | None = None, guard_ctrltime: ConvertibleExpression | None = None, guard_dist: ConvertibleExpression | None = None, yaccel: ConvertibleExpression | None = None, ground_velocity: TupleExpression | None = None, guard_velocity: ConvertibleExpression | None = None, air_velocity: TupleExpression | None = None, airguard_velocity: TupleExpression | None = None, ground_cornerpush_veloff: ConvertibleExpression | None = None, air_cornerpush_veloff: ConvertibleExpression | None = None, down_cornerpush_veloff: ConvertibleExpression | None = None, guard_cornerpush_veloff: ConvertibleExpression | None = None, airguard_cornerpush_veloff: ConvertibleExpression | None = None, airguard_ctrltime: ConvertibleExpression | None = None, air_juggle: ConvertibleExpression | None = None, mindist: TupleExpression | None = None, maxdist: TupleExpression | None = None, snap: TupleExpression | None = None, p1sprpriority: ConvertibleExpression | None = None, p2sprpriority: ConvertibleExpression | None = None, p1facing: ConvertibleExpression | None = None, p1getp2facing: ConvertibleExpression | None = None, p2facing: ConvertibleExpression | None = None, p1stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p2stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p2getp1state: ConvertibleExpression | None = None, forcestand: ConvertibleExpression | None = None, fall: ConvertibleExpression | None = None, fall_xvelocity: ConvertibleExpression | None = None, fall_yvelocity: ConvertibleExpression | None = None, fall_recover: ConvertibleExpression | None = None, fall_recovertime: ConvertibleExpression | None = None, fall_damage: ConvertibleExpression | None = None, air_fall: ConvertibleExpression | None = None, forcenofall: ConvertibleExpression | None = None, down_velocity: TupleExpression | None = None, down_hittime: ConvertibleExpression | None = None, down_bounce: ConvertibleExpression | None = None, chainid: ConvertibleExpression | None = None, nochainid: TupleExpression | None = None, hitonce: ConvertibleExpression | None = None, kill: ConvertibleExpression | None = None, guard_kill: ConvertibleExpression | None = None, fall_kill: ConvertibleExpression | None = None, numhits: ConvertibleExpression | None = None, getpower: TupleExpression | None = None, givepower: TupleExpression | None = None, palfx_time: ConvertibleExpression | None = None, palfx_mul: TupleExpression | None = None, palfx_add: TupleExpression | None = None, envshake_time: ConvertibleExpression | None = None, envshake_freq: ConvertibleExpression | None = None, envshake_ampl: ConvertibleExpression | None = None, envshake_phase: ConvertibleExpression | None = None, fall_envshake_time: ConvertibleExpression | None = None, fall_envshake_freq: ConvertibleExpression | None = None, fall_envshake_ampl: ConvertibleExpression | None = None, fall_envshake_phase: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitFallDamage(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitFallSet(value: ConvertibleExpression | None = None, xvel: ConvertibleExpression | None = None, yvel: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitFallVel(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitOverride(attr: TupleExpression, stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, slot: ConvertibleExpression | None = None, time: ConvertibleExpression | None = None, forceair: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def HitVelSet(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def LifeAdd(value: ConvertibleExpression, kill: ConvertibleExpression | None = None, absolute: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def LifeSet(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def MakeDust(pos: TupleExpression | None = None, pos2: TupleExpression | None = None, spacing: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ModifyExplod(id: ConvertibleExpression, pos: TupleExpression | None = None, postype: ConvertibleExpression | None = None, facing: ConvertibleExpression | None = None, vfacing: ConvertibleExpression | None = None, bindtime: ConvertibleExpression | None = None, vel: TupleExpression | None = None, accel: TupleExpression | None = None, random: TupleExpression | None = None, removetime: ConvertibleExpression | None = None, supermove: ConvertibleExpression | None = None, supermovetime: ConvertibleExpression | None = None, pausemovetime: ConvertibleExpression | None = None, scale: TupleExpression | None = None, sprpriority: ConvertibleExpression | None = None, ontop: ConvertibleExpression | None = None, shadow: ConvertibleExpression | None = None, ownpal: ConvertibleExpression | None = None, removeongethit: ConvertibleExpression | None = None, trans: ConvertibleExpression | None = None, alpha: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def MoveHitReset(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def NotHitBy(value: TupleExpression | None = None, value2: TupleExpression | None = None, time: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Null(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Offset(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PalFX(time: ConvertibleExpression | None = None, add: TupleExpression | None = None, mul: TupleExpression | None = None, sinadd: TupleExpression | None = None, invertall: ConvertibleExpression | None = None, color: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ParentVarAdd(var: ConvertibleExpression | None = None, value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ParentVarSet(var: ConvertibleExpression | None = None, value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Pause(time: ConvertibleExpression, endcmdbuftime: ConvertibleExpression | None = None, movetime: ConvertibleExpression | None = None, pausebg: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PlayerPush(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PlaySnd(value: TupleExpression, volumescale: ConvertibleExpression | None = None, channel: ConvertibleExpression | None = None, lowpriority: ConvertibleExpression | None = None, freqmul: ConvertibleExpression | None = None, loop: ConvertibleExpression | None = None, pan: ConvertibleExpression | None = None, abspan: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PosAdd(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PosFreeze(value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PosSet(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PowerAdd(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def PowerSet(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Projectile(attr: TupleExpression, projid: ConvertibleExpression | None = None, projanim: ConvertibleExpression | Animation | None = None, projhitanim: ConvertibleExpression | Animation | None = None, projremanim: ConvertibleExpression | Animation | None = None, projscale: TupleExpression | None = None, projremove: ConvertibleExpression | None = None, projremovetime: ConvertibleExpression | None = None, velocity: TupleExpression | None = None, remvelocity: TupleExpression | None = None, accel: TupleExpression | None = None, velmul: TupleExpression | None = None, projhits: ConvertibleExpression | None = None, projmisstime: ConvertibleExpression | None = None, projpriority: ConvertibleExpression | None = None, projedgebound: ConvertibleExpression | None = None, projstagebound: ConvertibleExpression | None = None, projheightbound: TupleExpression | None = None, offset: TupleExpression | None = None, postype: ConvertibleExpression | None = None, projshadow: ConvertibleExpression | None = None, supermovetime: ConvertibleExpression | None = None, pausemovetime: ConvertibleExpression | None = None, afterimage_time: ConvertibleExpression | None = None, afterimage_length: ConvertibleExpression | None = None, afterimage_palcolor: ConvertibleExpression | None = None, afterimage_palinvertall: ConvertibleExpression | None = None, afterimage_palbright: TupleExpression | None = None, afterimage_palcontrast: TupleExpression | None = None, afterimage_palpostbright: TupleExpression | None = None, afterimage_paladd: TupleExpression | None = None, afterimage_palmul: TupleExpression | None = None, afterimage_timegap: ConvertibleExpression | None = None, afterimage_framegap: ConvertibleExpression | None = None, afterimage_trans: ConvertibleExpression | None = None, hitflag: ConvertibleExpression | None = None, guardflag: ConvertibleExpression | None = None, affectteam: ConvertibleExpression | None = None, animtype: ConvertibleExpression | None = None, air_animtype: ConvertibleExpression | None = None, fall_animtype: ConvertibleExpression | None = None, priority: TupleExpression | None = None, damage: ConvertibleExpression | None = None, pausetime: ConvertibleExpression | None = None, guard_pausetime: ConvertibleExpression | None = None, sparkno: ConvertibleExpression | None = None, guard_sparkno: ConvertibleExpression | None = None, sparkxy: TupleExpression | None = None, hitsound: TupleExpression | None = None, guardsound: TupleExpression | None = None, ground_type: ConvertibleExpression | None = None, air_type: ConvertibleExpression | None = None, ground_slidetime: ConvertibleExpression | None = None, guard_slidetime: ConvertibleExpression | None = None, ground_hittime: ConvertibleExpression | None = None, guard_hittime: ConvertibleExpression | None = None, air_hittime: ConvertibleExpression | None = None, guard_ctrltime: ConvertibleExpression | None = None, guard_dist: ConvertibleExpression | None = None, yaccel: ConvertibleExpression | None = None, ground_velocity: TupleExpression | None = None, guard_velocity: TupleExpression | None = None, air_velocity: TupleExpression | None = None, airguard_velocity: TupleExpression | None = None, ground_cornerpush_veloff: ConvertibleExpression | None = None, air_cornerpush_veloff: ConvertibleExpression | None = None, down_cornerpush_veloff: ConvertibleExpression | None = None, guard_cornerpush_veloff: ConvertibleExpression | None = None, airguard_cornerpush_veloff: ConvertibleExpression | None = None, airguard_ctrltime: ConvertibleExpression | None = None, air_juggle: ConvertibleExpression | None = None, mindist: TupleExpression | None = None, maxdist: TupleExpression | None = None, snap: TupleExpression | None = None, p1sprpriority: ConvertibleExpression | None = None, p2sprpriority: ConvertibleExpression | None = None, p1facing: ConvertibleExpression | None = None, p1getp2facing: ConvertibleExpression | None = None, p2facing: ConvertibleExpression | None = None, p1stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p2stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p2getp1state: ConvertibleExpression | None = None, forcestand: ConvertibleExpression | None = None, fall: ConvertibleExpression | None = None, fall_xvelocity: ConvertibleExpression | None = None, fall_yvelocity: ConvertibleExpression | None = None, fall_recover: ConvertibleExpression | None = None, fall_recovertime: ConvertibleExpression | None = None, fall_damage: ConvertibleExpression | None = None, air_fall: ConvertibleExpression | None = None, forcenofall: ConvertibleExpression | None = None, down_velocity: TupleExpression | None = None, down_hittime: ConvertibleExpression | None = None, down_bounce: ConvertibleExpression | None = None, chainid: ConvertibleExpression | None = None, nochainid: TupleExpression | None = None, hitonce: ConvertibleExpression | None = None, kill: ConvertibleExpression | None = None, guard_kill: ConvertibleExpression | None = None, fall_kill: ConvertibleExpression | None = None, numhits: ConvertibleExpression | None = None, getpower: TupleExpression | None = None, givepower: TupleExpression | None = None, palfx_time: ConvertibleExpression | None = None, palfx_mul: TupleExpression | None = None, palfx_add: TupleExpression | None = None, envshake_time: ConvertibleExpression | None = None, envshake_freq: ConvertibleExpression | None = None, envshake_ampl: ConvertibleExpression | None = None, envshake_phase: ConvertibleExpression | None = None, fall_envshake_time: ConvertibleExpression | None = None, fall_envshake_freq: ConvertibleExpression | None = None, fall_envshake_ampl: ConvertibleExpression | None = None, fall_envshake_phase: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def RemapPal(source: TupleExpression, dest: TupleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def RemoveExplod(id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ReversalDef(reversal_attr: TupleExpression, hitflag: ConvertibleExpression | None = None, pausetime: TupleExpression | None = None, sparkno: ConvertibleExpression | None = None, guard_sparkno: ConvertibleExpression | None = None, hitsound: TupleExpression | None = None, guardsound: TupleExpression | None = None, p1stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p2stateno: ConvertibleExpression | str | int | Callable[..., None | StateController] | None = None, p1sprpriority: ConvertibleExpression | None = None, p2sprpriority: ConvertibleExpression | None = None, sparkxy: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def ScreenBound(value: ConvertibleExpression | None = None, movecamera: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def SelfState(value: ConvertibleExpression, ctrl: ConvertibleExpression | None = None, anim: ConvertibleExpression | Animation | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def SprPriority(value: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def StateTypeSet(statetype: ConvertibleExpression | None = None, movetype: ConvertibleExpression | None = None, physics: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def SndPan(channel: ConvertibleExpression, pan: ConvertibleExpression, abspan: ConvertibleExpression, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def SuperPause(time: ConvertibleExpression | None = None, anim: ConvertibleExpression | None = None, sound: TupleExpression | None = None, pos: TupleExpression | None = None, darken: ConvertibleExpression | None = None, p2defmul: ConvertibleExpression | None = None, poweradd: ConvertibleExpression | None = None, unhittable: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetBind(time: ConvertibleExpression | None = None, id: ConvertibleExpression | None = None, pos: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetDrop(excludeid: ConvertibleExpression | None = None, keepone: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetFacing(value: ConvertibleExpression, id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetLifeAdd(value: ConvertibleExpression, id: ConvertibleExpression | None = None, kill: ConvertibleExpression | None = None, absolute: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetPowerAdd(value: ConvertibleExpression, id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetState(value: ConvertibleExpression | str | int | Callable[..., None | StateController], id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetVelAdd(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def TargetVelSet(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, id: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Trans(trans: ConvertibleExpression, alpha: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Turn(ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VarAdd(var: ConvertibleExpression | None = None, value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VarSet(var: ConvertibleExpression | None = None, value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VarRandom(v: ConvertibleExpression, range: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VarRangeSet(value: ConvertibleExpression, fvalue: ConvertibleExpression, first: ConvertibleExpression | None = None, last: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VelAdd(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VelMul(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VelSet(x: ConvertibleExpression | None = None, y: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def VictoryQuote(value: ConvertibleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
def Width(edge: TupleExpression | None = None, player: TupleExpression | None = None, value: TupleExpression | None = None, ignorehitpause: ConvertibleExpression | None = None, persistent: ConvertibleExpression | None = None) -> StateController: ...
